part of tecfy_core_client_package;

class LoginController extends BaseController {
  bool isLocked = false;

  /// form key
  late GlobalKey formKey;

  /// username text controller
  late TextEditingController usernameController;

  /// password text controller
  late TextEditingController passwordController;

  /// account service
  late AccountService service;

  /// user name validator
  String usernameValidator(String? v) =>
      v == null || v.isEmpty ? 'Please Enter your Email or Username' : '';

  /// password validator
  String passwordValidator(String? v) =>
      v == null || v.isEmpty ? 'Please Enter your Password' : '';

  /// login page controller
  LoginController(State<StatefulWidget> state) : super(state) {
    formKey = GlobalKey<FormState>();
    usernameController = TextEditingController();
    passwordController = TextEditingController();
    service = AccountService();
    isLocked = ConfigService.getValueBool('LoginExpired');
    usernameController.text = LocalConfigService.user?.username ?? '';
  }

  Future onLogout() {
    return AccountService.signout(context);
  }

  /// login and goto main page
  Future onLogin() async {
    var result = await service.login(
        usernameController.text.trim(), passwordController.text.trim());
    print('================== Login result============');
    print(result);
    print(result.success);
    print(result.message);
    if (result.success) {
      await AccountService.init(context);
    } else {
      if (result.errorCodeString == 'ClientVersionNotSupported') {
        TecfyCoreApp.mustUpgrade = true;
        AppLocale.languageStreamController.add(LocalConfigService.language);
        return;
      }
      showMessage(result.message ?? "Error", type: MessageType.error);
    }
  }

  /// show contact us page
  contactUs() async {
    Widget getField(label, value) {
      return Row(crossAxisAlignment: CrossAxisAlignment.start, children: [
        Text('${AppLocale.translate(context, label)}:',
            style: TextStyle(fontWeight: FontWeight.bold)),
        SizedBox(width: 10),
        Expanded(child: Text(value, maxLines: 5)),
      ]);
    }

    await AppNavigator.showAppDialog(
      context,
      (c) => Material(
        type: MaterialType.transparency,
        child: Center(
          child: Container(
            width: 300,
            height: 300,
            child: Card(
                margin: EdgeInsets.all(10),
                child: Padding(
                    padding: EdgeInsets.all(10),
                    child: Column(children: [
                      getField('Website', 'https://tecfy.co'),
                      getField('Email', 'info@tecfy.co'),
                      getField('Address',
                          'Office 3 Building 3 on Hafez Ibrahem and Babel Street, Alsharq District, Port Said, Egypt'),
                      Expanded(child: Container()),
                      TextButton(
                          child: Text(AppLocale.translate(context, 'Close'),
                              style: TextStyle(color: ThemeService.subColor)),
                          onPressed: () => Navigator.of(context).pop())
                    ]))),
          ),
        ),
      ),
    );
  }
}
