part of tecfy_core_client_package;

class TecfyCoreApp extends StatelessWidget {
  /// used to redirect to must upgrade page
  static bool mustUpgrade = false;

  /// app Global key used for navigate
  static GlobalKey<NavigatorState> appGlobalkey = GlobalKey<NavigatorState>();

  /// current version number
  static String get currentVersion => TecfyBasicApp.appVersion;
  static set currentVersion(String v) => TecfyBasicApp.appVersion = v;

  /// API rul
  static set apiURL(v) => ConfigService.apiURL = v;

  /// API URL
  static set wsURL(v) => ConfigService.wsURL = v;

  /// entities service
  static Map<String, IEntityTransaction> entityServices =
      Map<String, IEntityTransaction>();

  /// entities action service
  static Map<String, IEntityAction> entityActionsServices =
      Map<String, IEntityAction>();

  /// them config
  static set themeConfig(IThemeConfig? v) => TecfyBasicApp.themeConfig = v;

  /// them config
  static IThemeConfig? get themeConfig => TecfyBasicApp.themeConfig;

  /// app logo
  static Image? logo;

  /// image
  static late Image backgoundImage;

  /// translate service for multi language
  static late ITranslateService translateService;

  /// report service
  static late IReportService reportService;

  /// navigate to main page after login
  static FutureOr Function(BuildContext) navigateToDefaultPage = (c) =>
      AppNavigator.navigateToNamed(c, '/',
          replaceAll: true, animationType: AnimationTypes.opacity);

  /// update local db function
  /// this function called after login
  static Future Function(Database, int, int) updateSqlLiteDb =
      (d, f, t) async {};

  /// db version
  static int appDbVersion = 0;

  /// menu service
  static IMenuInterface? menuService;

  /// package info
  static late PackageInfo packageInfo;
  // This widget is the root of your application.
  @override
  Widget build(BuildContext context) {
    // kReleaseMode;
    if (kReleaseMode) {
      return StreamBuilder<Language>(
          stream: AppLocale.languageStream,
          initialData: LocalConfigService.language,
          builder: (context, languageSnapshot) {
            AppLocale.appLocale = null;
            return _getApp(languageSnapshot.data ?? Language.arabic);
          });
    } else {
      return _getApp(LocalConfigService.language);
    }
  }

  _getApp(Language language) {
    return MaterialApp(
      key: kReleaseMode ? UniqueKey() : null,
      title: 'Application',
      navigatorKey: appGlobalkey,
      builder: BotToastInit(),
      theme: ThemeService.getMainThemeAndroid(),
      supportedLocales: [const Locale('ar', 'SA'), const Locale('en', 'US')],
      localizationsDelegates: [
        const AppLocalizationsDelegate(),
        GlobalCupertinoLocalizations.delegate,
        DefaultCupertinoLocalizations.delegate,
        GlobalMaterialLocalizations.delegate,
        GlobalWidgetsLocalizations.delegate,
        DefaultCupertinoLocalizations.delegate
      ],
      locale: language == Language.arabic
          ? const Locale('ar', 'SA')
          : const Locale('en', 'US'),
      localeResolutionCallback:
          (Locale? deviceLocale, Iterable<Locale> supportedLocales) {
        ConfigService.isRTL = (language == Language.arabic);
        var selectedLanguageCode = (language == Language.arabic) ? 'ar' : 'en';
        var lcl = supportedLocales
            .firstWhere((d) => d.languageCode == selectedLanguageCode);

        return lcl;
      },
      //home: SplashScreenPage(),
      onGenerateRoute: (RouteSettings settings) {
        if (mustUpgrade)
          return GeneratePageRoute(
              widget: UpgradePage(), routeName: '/upgrade');

        if (!InitService.isInitled)
          return GeneratePageRoute(
              widget: SplashScreenPage(), routeName: settings.name ?? '');

        switch (settings.name) {
          case '/':
            return GeneratePageRoute(
                widget: DashboardPage(), routeName: settings.name ?? '');
          case '/loading':
            return GeneratePageRoute(
                widget: SplashScreenPage(), routeName: settings.name ?? '');
          case '/login':
            return GeneratePageRoute(
                widget: LoginPage(), routeName: settings.name ?? '');
          case '/setting':
            return GeneratePageRoute(
                widget: SettingPage(), routeName: settings.name ?? '');
          case '/upgrade':
            return GeneratePageRoute(
                widget: UpgradePage(), routeName: settings.name ?? '');
          default:
            return GeneratePageRoute(
                widget: DashboardPage(), routeName: settings.name ?? '');
        }
      },
      initialRoute: mustUpgrade ? '/upgrade' : '/loading',
    );
  }

  /// init application
  static void init({String? mapAPI}) {
    if (!kIsWeb && (Platform.isWindows || Platform.isLinux)) {
      // Initialize FFI
      sqfliteFfiInit();
      // Change the default factory
      databaseFactory = databaseFactoryFfi;
    }
    if (mapAPI != null) {
      //GoogleMap.init(mapAPI);
    }
    IApiService.unauthrizedCallback = (response) {
      Future.delayed(Duration(milliseconds: 10)).then((v) {
        AccountService.loginExpired(appGlobalkey.currentContext);
      });
    };
  }
}

class GeneratePageRoute extends PageRouteBuilder {
  /// widget
  final Widget widget;

  /// route name for url navigator
  final String routeName;

  /// rounting widget to restart the app after language changes
  GeneratePageRoute({required this.widget, required this.routeName})
      : super(
          settings: RouteSettings(name: routeName),
          pageBuilder: (BuildContext context, Animation<double> animation,
              Animation<double> secondaryAnimation) {
            return widget;
          },
        );
}
