part of tecfy_core_client_package;

class Message {
  late String id;
  late String fromUserId;
  late String fromName;

  late String toUserId;
  late String toName;

  late String body;
  late DateTime? time;
  String get timeString => time == null
      ? '-'
      : time!.difference(DateTime.now().toUtc()).inDays <= 0
          ? intl.DateFormat("hh:mm a").format(time!.toLocal())
          : intl.DateFormat("yyyy-MM-dd hh:mm a").format(time!.toLocal());

  bool imSender = false;

  Message(
      {required this.id,
      required this.body,
      required this.fromName,
      required this.toName,
      required this.fromUserId,
      required this.toUserId,
      required this.imSender,
      required this.time});

  Message.sendNew({required this.body, required this.toUserId}) {
    id = Uuid().v4().replaceAll('-', '');
    time = DateTime.now().toUtc();
    fromUserId = LocalConfigService.user!.id;
    imSender = true;
  }

  Message.fromMap(Map<String, dynamic> row) {
    try {
      id = row['id'];
      fromUserId = row['fromUserId'];
      fromName = row['fromName'];
      toUserId = row['toUserId'];
      toName = row['toName'];
      body = row['body'];
      //imSender = (row['fromUserId'] == LocalConfigService.user.id);
      time = BaseModel.parseUTC(row['time']) ?? DateTime.now();
      //time = DateTime.tryParse(row['time'].toString().substring(0,23).replaceFirst('T', ' '));
    } catch (err) {
      print('---------------------------');
      print(row);
      print(err);
    }
  }

  Message.fromMapHistory(Map<String, dynamic> row) {
    try {
      id = row['id'];
      fromUserId = row['fromUserId'];
      fromName = row['fromName'];
      toUserId = row['toUserId'];
      toName = row['toName'];
      body = row['body'];
      imSender = (row['fromUserId'] == LocalConfigService.user!.id);
      time = DateTime.tryParse(
              row['time'].toString().substring(0, 23).replaceFirst('T', ' ')) ??
          DateTime.now();
    } catch (err) {
      print('---------------------------');
      print(row);
      print(err);
    }
  }

  Message.fromMapDatabase(Map<String, dynamic> row) {
    try {
      id = row['id'];
      fromUserId = row['fromUserId'];
      fromName = row['fromName'];
      toUserId = row['toUserId'];
      toName = row['toName'];
      body = row['body'];
      imSender = row['imSender'] == 1;
      time = DateTime.fromMillisecondsSinceEpoch(row['time'] ?? 0, isUtc: true);
    } catch (err) {
      print('--------------------------- Error on fromMapDatabase');
      print(json.encode(row));
      print(err);
    }
  }

  toMap() {
    return {'toUserId': toUserId, 'body': body, 'id': id};
  }

  toMapDb() {
    return {
      'id': id,
      'fromUserId': fromUserId,
      'fromName': fromName,
      'toUserId': toUserId,
      'toName': toName,
      'body': body,
      'imSender': imSender ? 1 : 0,
      'time': time!.millisecondsSinceEpoch
    };
  }

  String get createdSince {
    try {
      if (time == null) return 'just now';

      var differance = DateTime.now().toUtc().difference(time!);
      var years = (differance.inDays ~/ 365).toInt();
      if (years > 1) return '$years years ago';

      var months = (differance.inDays ~/ 30).toInt();
      if (months > 1) return '$months months ago';

      var weeks = (differance.inDays ~/ 7).toInt();
      if (weeks > 1) return '$weeks weeks ago';

      var days = (differance.inDays).toInt();
      if (days > 1) return '$days days ago';

      var hours = (differance.inHours).toInt();
      if (hours > 1) return '$hours hours ago';

      var minutes = (differance.inMinutes).toInt();
      if (minutes > 1) return '$minutes minutes ago';
    } catch (err) {}
    return 'Just now';
  }
}
