part of tecfy_core_client_package;

class DatabaseService {
  static Database? _db;
  static String? pathString;
  static String? _dbPath = 'main.db';
  static LocalStorage? storage;
  static bool isInit = false;
  static bool isEntityTableCreated = false;
  static bool get isLocalDbEnabled => _db != null;
  static List<DbTransaction>? trans;
  static bool supportOnConflict = false;

  static Future init() async {
    try {
      if (storage == null) storage = new LocalStorage('localDB.json');
    } catch (err) {}
    isInit = true;
    if (_db != null && (_db?.isOpen ?? false)) {
      await _createDatabaseEntityTable(_db);
      return;
    }
    if (!ConfigService.isLogin && !TecfyBasicApp.isTest) return;
    try {
      if (PlatformService.isSupportLocalDBSupported) {
        _dbPath = '${LocalConfigService.user!.id}.tfd';
        pathString =
            '${(await path.getApplicationDocumentsDirectory()).path}${Platform.isWindows ? '\\' : '/'}';
        print('Database Path $pathString   $_dbPath');
        //await deleteDatabase(_dbPath);
        if (TecfyBasicApp.isTest) {
          _db = await databaseFactoryFfi.openDatabase(inMemoryDatabasePath);
          await _createDatabase(_db!, 0, 0);
        } else
          _db = await openDatabase(
              Platform.isWindows ? '$pathString$_dbPath' : _dbPath!,
              version: 2,
              onCreate: (db, v) => _createDatabase(db, 0, v),
              onUpgrade: _createDatabase,
              onDowngrade: _createDatabase);
        try {
          await _db?.execute(
              "insert into Notification (id) values ('0') on conflict(id) DO UPDATE SET titleAr = 'TEST'");
          supportOnConflict = true;
          await _db?.execute("delete from Notification where id = '0'");
        } catch (err) {
          print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX NOT supportOnConflict');
          print(err);
        }
        print('db created');
        await _createDatabaseEntityTable(_db);
        print('db has been inited successfully !!!!!!!!!!!!!');
      } else {
        pathString = "d:\\";
      }
    } catch (err) {
      // cannot get path means it's windows
      pathString = "d:\\";
      print('xxxxxxxxxxxxxx  error inital database xxxxxxxxxxxx');
      print(err);
    }
  }

  static Future<OperationResult> exeBatch(
      Future<OperationResult> Function(Batch) process,
      [bool continueOnError = false]) async {
    try {
      print('Start Execute batch -------------');
      if (_db == null)
        return OperationResult(
            success: false, message: 'قاعده البيانات المحليه غير موجوده');
      var batch = _db?.batch();
      var result = await process(batch!);
      if (result.success) {
        batch.commit(continueOnError: continueOnError);
      }
      return result;
    } catch (err, t) {
      print('XXXXXXXXXXXXXX error executing batch:');
      print(err);
      print(t);
      return OperationResult(
          success: false, message: err.toString(), innerError: t.toString());
    }
  }

  static Future<List<Map<String, Object?>>?> rawQuery(String sql,
      [List<Object?>? arguments]) async {
    if (_db == null || !(_db?.isOpen ?? false)) {
      return Future.value([]);
    }
    return await _db?.rawQuery(sql, arguments);
  }

  static Future<int> insert(String table, Map<String, dynamic> values,
      {String? nullColumnHack, ConflictAlgorithm? conflictAlgorithm}) async {
    if (_db == null || !(_db?.isOpen ?? false)) {
      return 0;
    }
    await _db?.insert(
        table, toDatabaseEntity(EntityService.getEntity(table), values, false)!,
        nullColumnHack: nullColumnHack, conflictAlgorithm: conflictAlgorithm);
    return 10;
  }

  static Future<int?> delete(String table,
      {String? where, List<dynamic>? whereArgs}) async {
    if (_db == null || !(_db?.isOpen ?? false)) {
      return Future.value(0);
    }
    return await _db?.delete(table, where: where, whereArgs: whereArgs);
  }

  static Future<int?> update(String table, Map<String, dynamic> values,
      {String? where,
      List<dynamic>? whereArgs,
      ConflictAlgorithm? conflictAlgorithm,
      bool skipDataUpdate = false}) async {
    if (_db == null || !(_db?.isOpen ?? false)) {
      return Future.value(0);
    }
    return await _db?.update(
        table,
        (skipDataUpdate
                ? values
                : toDatabaseEntity(
                    EntityService.getEntity(table), values, true)) ??
            {},
        where: where,
        whereArgs: whereArgs,
        conflictAlgorithm: conflictAlgorithm);
  }

  static Future<List<Map<String, dynamic>>?> query(String table,
      {bool? distinct,
      List<String>? columns,
      String? where,
      List<dynamic>? whereArgs,
      String? groupBy,
      String? having,
      String? orderBy,
      int? limit,
      int? offset}) async {
    if (_db == null || !(_db?.isOpen ?? false)) {
      return Future.value([]);
    }
    return await _db?.query(table,
        distinct: distinct,
        columns: columns,
        where: where,
        whereArgs: whereArgs,
        groupBy: groupBy,
        limit: limit,
        orderBy: orderBy,
        offset: offset);
  }

  static Future<void> _createDatabase(
      Database _db, int _fromVerison, int _toVersion) async {
    try {
      // var fmVersion =0;
      // var appVersion =0;
      //if (_fromVerison <= 1) {
      await _db.execute(
          'CREATE TABLE IF NOT EXISTS Message (id TEXT PRIMARY KEY, fromUserId TEXT,toUserId TEXT,body TEXT, fromName TEXT,toName TEXT, time INTEGER,imSender INTEGER)');
      await _db.execute(
          'CREATE TABLE IF NOT EXISTS Notification (id TEXT PRIMARY KEY, titleAr TEXT,titleEn TEXT,messageAr TEXT, messageEn TEXT, time INTEGER, seen INTEGER, type INTEGER, payload TEXT)');
      await _db.execute(
          'CREATE TABLE IF NOT EXISTS Entity (id TEXT PRIMARY KEY, entity TEXT,body TEXT, timeStamp INTEGER, recordNumber INTEGER,isPosted INTEGER,postMessage TEXT)');
      await _db.execute(
          'CREATE TABLE IF NOT EXISTS DeleteLog (id TEXT PRIMARY KEY, entity TEXT, timeStamp INTEGER,isPosted INTEGER,postMessage TEXT)');

      await _db.execute(
          'CREATE INDEX IF NOT EXISTS  INX_Entity_entity ON Entity(entity,timeStamp)');
      await _db.execute(
          'CREATE INDEX IF NOT EXISTS  INX_Entity_isPosted ON Entity(isPosted)');
      await _db.execute(
          'CREATE INDEX IF NOT EXISTS  INX_DeleteLog_isPosted ON DeleteLog(isPosted)');

      //}
      await TecfyCoreApp.updateSqlLiteDb(_db, _fromVerison, _toVersion);
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXXXX databse create builtin tables');
      print(err);
    }
  }

  static List<String> builtInFields = [
    'id',
    'isPosted',
    'postMessage',
    'timeStamp',
    'recordNumber'
  ];

  static Future<void> updateEntityTable() {
    return _createDatabaseEntityTable(_db);
  }

  static Future<void> _createDatabaseEntityTable(Database? _db) async {
    try {
      if (EntityService.entities == null) {
        print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX no entiotoies');
        return;
      }
      print('--------Start create/update entity Table');
      for (var entity in EntityService.entities!.values) {
        var existingTable =
            await _db?.rawQuery('PRAGMA table_info(${entity.dbName})');

        if (existingTable != null && existingTable.length > 0) {
          for (var prop in entity.properties) {
            if (builtInFields.contains(prop.dbName)) {
              continue;
            }
            var type = 'TEXT';
            var propName = prop.dbName;
            if (prop.propertyType == PropertyTypes.File) {
              propName = '${prop.dbName}Id';
              type = 'TEXT';
            } else {
              switch (prop.valueType) {
                case ValueTypes.DateTime:
                  type = 'INTEGER';
                  break;
                case ValueTypes.Boolean:
                  type = 'INTEGER';
                  break;
                case ValueTypes.Number:
                  type = 'INTEGER';
                  break;
                default:
              }
            }
            // cols.add('${prop.dbName} $type');
            if (existingTable.any((e) => e['name'] == propName)) {
              if (existingTable
                      .firstWhere((e) => e['name'] == propName)['type'] !=
                  type) {
                print(
                    '||||||||||||||||||||||||||| update type ${entity.dbName}=>$propName $type');
              }
            } else {
              print(
                  '||||||||||||||||||||||||||||| add column ${entity.dbName}=>$propName $type');
              try {
                await _db?.execute(
                    'ALTER TABLE ${entity.dbName} add $propName $type');
              } catch (err) {
                print('XXXXX add column error : $err');
              }
            }
          } // property loop

          // if(existingTable.any((e)=>e['name'] == 'isPosted'))

        } else {
          print(' table not exist, start create ${entity.dbName}');
          List<String> cols = [
            'id TEXT PRIMARY KEY',
            'isPosted INTEGER',
            'recordNumber INTEGER',
            'postMessage TEXT',
            'timeStamp INTEGER'
          ];

          for (var prop in entity.properties) {
            if (builtInFields.contains(prop.dbName)) {
              continue;
            }
            if (prop.propertyType == PropertyTypes.File) {
              cols.add('${prop.dbName}Id TEXT');
              continue;
            }
            var type = 'TEXT';
            switch (prop.valueType) {
              case ValueTypes.DateTime:
                type = 'INTEGER';
                break;
              case ValueTypes.Boolean:
                type = 'INTEGER';
                break;
              case ValueTypes.Number:
                type = 'INTEGER';
                break;
              // case ValueTypes.NumberArray:
              //   type = 'TEXT';
              //   break;
              default:
            }
            cols.add('${prop.dbName} $type');
          }
          var sql =
              'CREATE TABLE IF NOT EXISTS ${entity.dbName} (${cols.join(', ')})';
          print(sql);
          await _db?.execute(sql);
        }
      }
      print(
          '------------------- all entity tables has been created ${EntityService.entities!.values.length}');
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX _createDatabaseEntityTable');
      print(err);
    }
  }

  static Future<void> deleteDatabaseFile() async {
    try {
      _db?.close();
      await deleteDatabase(_dbPath ?? '');
    } catch (err) {
      print('xxxxxxxxxxxxxx  error deleting database xxxxxxxxxxxx');
      print(err);
    }
  }

  static Future<bool> removeRange(String entity, List<EntityFilter> filters,
      {bool withTrans = false, String divider = 'and'}) async {
    var w = DatabaseService.getWhere(filters);

    if (withTrans)
      trans?.add(DbTransaction(QueryType.delete,
          table: entity,
          where: w['where'].join(' $divider '),
          whereArgs: w['whereArgs']));
    else
      await _db?.delete(entity,
          where: w['where'].join(' and '), whereArgs: w['whereArgs']);
    return true;
  }

  static dynamic getWhere(List<EntityFilter> filters) {
    var where = [];
    var whereArgs = [];

    for (var f in filters) {
      if (f.value == null) {
        where.add('${f.property?.dbName ?? f.propertyName} is null');
      } else {
        var op = '=';
        if (f.operatorId == 'GE')
          op = '>=';
        else if (f.operatorId == 'LE')
          op = '<=';
        else if (f.operatorId == 'G')
          op = '>';
        else if (f.operatorId == 'C')
          op = 'like';
        else if (f.operatorId == 'NC')
          op = 'not like';
        else if (f.operatorId == 'L') op = '<';
        where.add('${f.property?.dbName ?? f.propertyName} $op ?');
        if (op == 'like' || op == 'not like')
          whereArgs.add('%${f.value}%');
        else
          whereArgs.add(f.value);
      }
    }
    return {'where': where, 'whereArgs': whereArgs};
  }

  static Future<bool> saveEntity(
      String entityName, List<Map<String, dynamic>> data,
      {bool isPosted = true,
      bool overwriteLocal = false,
      bool withTrans = false}) async {
    if (entityName.startsWith('_')) {
      var old = (await readCollection(entityName)).data ?? [];
      old.removeWhere((e) => data.any((d) => e['id'] == d['id']));
      old.addAll(data);
      return (await writeCollection(entityName, old)).success;
    }
    if (!ConnectionService.isOfflineMode) return false;
    // print(
    //     'Start ----------------------- entity save $entityName ${data.length}');

    var ids = data.map<String>((d) => d['id']).join("','");

    if (!withTrans) trans = [];

    trans?.add(DbTransaction(QueryType.delete,
        table: entityName,
        where: "${overwriteLocal ? '' : 'isPosted = 1 and '} id in ('$ids')"));

    var result = await addEntityAll(entityName, data,
        isPosted: isPosted, withTrans: true);
    if (!result.success || withTrans) return result.success;

    return await execTrans();
  }

  static Future<OperationResult> addEntityAll(
      String entityName, List<dynamic> data,
      {bool isPosted = true, bool withTrans = false}) async {
    if (!ConnectionService.isOfflineMode)
      return OperationResult(success: false, message: 'النظام يعمل اونلاين');
    try {
      var entity = EntityService.getEntity(entityName);
      print('--------- addEntityAll $entityName ${data.length} items ');
      List<Map<String, dynamic>> queryList = [];
      for (var d in data) {
        var newData = toDatabaseEntity(entity, d, false);
        if (newData == null) continue;
        newData['isPosted'] = isPosted ? 1 : 0;
        if (supportOnConflict) {
          var query =
              "INSERT INTO $entityName (${newData.keys.join(',')}) VALUES(${newData.values.map((dd) => '?').join(',')}) ON CONFLICT(id) DO UPDATE SET ${newData.keys.where((k) => k != 'id').map((k) => '$k=excluded.$k').join(', ')};";
          queryList.add({'query': query, 'args': newData.values.toList()});
        } else {
          var queryD = "Delete From $entityName where id = ?;";
          queryList.add({
            'query': queryD,
            'args': [newData['id']]
          });
          var query =
              "INSERT INTO $entityName (${newData.keys.join(',')}) VALUES(${newData.values.map((dd) => '?').join(',')});";
          queryList.add({'query': query, 'args': newData.values.toList()});
        }
      }
      if (withTrans) {
        for (var q in queryList) {
          trans?.add(DbTransaction(QueryType.rawQuery,
              query: q['query'], whereArgs: q['args']));
        }
        return OperationResult(success: true);
      } else {
        var result = await DatabaseService.exeBatch((batch) async {
          try {
            for (var q in queryList) {
              batch.execute(q['query'], q['args']);
            }
            await batch.commit();
            return OperationResult(success: true);
          } catch (err) {
            print('XXXX exec batch error : $err');
            return OperationResult(success: false, message: err.toString());
          }
        });
        return result;
      }
      // print(
      //     '---------------- entity saved $entityName ${dataList.length} $result');
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX addEntityAll');
      print(err);
      return OperationResult(success: false, message: err.toString());
    }
  }

  static Future<bool> saveEntitySingle(
      String entityName, Map<String, dynamic>? d,
      {bool isPosted = true, bool withTrans: false}) async {
    if (d == null) return false;
    if (entityName.startsWith('_')) {
      var old = (await readCollection(entityName)).data ?? [];
      old.removeWhere((e) => e['id'] == d['id']);
      old.add(d);
      return (await writeCollection(entityName, old)).success;
    }

    if (!ConnectionService.isOfflineMode) return false;
    // print('Start ----------------------- entity single save $entityName ');
    //var entity = EntityService.getEntity(entityName);
    var newData = d;
    newData['isPosted'] = isPosted ? 1 : 0;
    var isExistResult = d['id'] == null
        ? [
            {'count': 0}
          ]
        : await DatabaseService.rawQuery(
            'select count(*) from $entityName where id = ?', [d['id']]);
    bool isNew = false;
    if (isExistResult == null ||
        isExistResult.length == 0 ||
        isExistResult.first.values.first == 0) {
      isNew = true;
    }
    List<DbTransaction> t = [];
    if (isNew)
      t.add(DbTransaction(QueryType.insert, table: entityName, data: newData));
    else
      t.add(DbTransaction(QueryType.update,
          table: entityName,
          data: newData,
          where: 'id = ?',
          whereArgs: [d['id']]));

    bool result = true;
    if (withTrans) {
      trans?.addAll(t);
    } else {
      result = await execTrans(t);
    }

    return result;
  }

  static Future<int?> getCounts(String entity,
      {String where = '', List<dynamic>? whereArgs}) async {
    try {
      if (entity.startsWith('_')) {
        return (await readCollection('_ServerConnections')).data?.length;
      }
      //var entityStructure = EntityService.getEntity(entity);
      var q =
          'select count(*) from $entity ${where.isNotEmpty && !where.startsWith('where') ? 'where' : ''} $where';

      var result = await DatabaseService.rawQuery(q, whereArgs);
      if (result?.length == 0) return -1;
      return result?.first.values.first as int?;
    } catch (err) {
      print('XXXXXXXXXXXXX DatabaseService.getCounts');
      print(err);
      return null;
    }
  }

  static Future<List<Map<String, dynamic>>?> getEntities(String entity,
      {String where = '',
      List<dynamic>? whereArgs,
      String? orderBy = 'timeStamp',
      int? offset,
      int? limit}) async {
    try {
      if (entity.startsWith('_')) {
        return (await readCollection('_ServerConnections')).data;
      }
      // print('----local query----');
      // print(where);
      // print(whereArgs.map((a) => getDatabaseValue(a)).toList());
      var entityStructure = EntityService.getEntity(entity);
      if (entityStructure == null) return null;
      var items = (await DatabaseService.query(entity,
              where: where.isEmpty ? null : where,
              whereArgs: whereArgs?.map((a) => getDatabaseValue(a)).toList(),
              orderBy: orderBy,
              offset: offset,
              limit: limit))
          ?.map<Map<String, dynamic>>(
              (d) => fromDatabaseEntity(entityStructure, d))
          .toList();
      print(
          '----------------------------- get $entity from local DB (${items?.length})');

      return items;
    } catch (err) {
      print('XXXXXXXXXXXXX DatabaseService.getEntities');
      print(err);
      return null;
    }
  }

  static Future<int?> getEntitiesCount(String entity,
      {String where = '',
      List<dynamic>? whereArgs,
      String orderBy = 'timeStamp',
      int? offset,
      int? limit}) async {
    try {
      if (entity.startsWith('_')) {
        return (await readCollection('_ServerConnections')).data?.length;
      }
      var entityStructure = EntityService.getEntity(entity);
      var items = (await DatabaseService.rawQuery(
        where.isEmpty
            ? 'select count(*) from $entity'
            : 'select count(*) from $entity where $where',
        //where: where.isEmpty ? null : where,
        whereArgs?.map((a) => getDatabaseValue(a)).toList(),
      ))
          ?.map<Map<String, dynamic>>(
              (d) => fromDatabaseEntity(entityStructure!, d))
          .toList();
      print(
          '----------------------------- get $entity from local DB (${items?.length})');

      return items?.first.values.first ?? 0;
    } catch (err) {
      print('XXXXXXXXXXXXX DatabaseService.getEntities');
      print(err);
      return null;
    }
  }

  static Future<dynamic> getEntityById(String entity, String id) async {
    if (!ConnectionService.isOfflineMode) return null;
    var items =
        (await DatabaseService.query(entity, where: 'id = ?', whereArgs: [id]))
                //.map((d) => jsonDecode(d['body']))
                ?.toList() ??
            [];
    var ent = EntityService.getEntity(entity);
    if (ent == null) {
      print('XXXXXXXXXXXXXXXX not found the entity');
    }
    return items.length > 0 ? fromDatabaseEntity(ent!, items[0]) : null;
  }

  static Future<OperationResult> deleteEntityById(String entity, String id,
      {bool withTrans = false}) async {
    if (entity.startsWith('_')) {
      var old = (await readCollection(entity)).data ?? [];
      old.removeWhere((e) => e['id'] == id);
      return (await writeCollection(entity, old));
    }
    if (!ConnectionService.isOfflineMode) return OperationResult(success: true);
    List<DbTransaction> t = [];
    t.add(DbTransaction(QueryType.rawQuery,
        query: 'delete from DeleteLog where isPosted = 1'));
    t.add(DbTransaction(QueryType.delete,
        table: 'DeleteLog', where: 'id = ?', whereArgs: [id]));

    t.add(DbTransaction(QueryType.insert, table: 'DeleteLog', data: {
      'id': id,
      'timeStamp': DateTime.now().millisecondsSinceEpoch,
      'entity': entity,
      'postMessage': '',
      'isPosted': 0
    }));
    t.add(DbTransaction(QueryType.delete,
        table: entity, where: 'id = ?', whereArgs: [id]));
    bool result = true;
    if (withTrans) {
      trans!.addAll(t);
    } else {
      result = await execTrans(t);
    }

    return OperationResult(success: result);
  }

  static Future<bool> execTrans([List<DbTransaction>? _trans]) async {
    // print('|||||||||||||||||||||||||||||| Save DB Trans |||||||||||');
    List<DbTransaction>? transactions = _trans ?? trans;
    // print(transactions?.length ?? -1);
    if (transactions == null || transactions.length == 0) {
      print('no trans to execute!');
      return true;
    }
    try {
      if (_db == null) await init();
      await _db!.transaction((txn) async {
        for (DbTransaction t in transactions) {
          //print(t);
          dynamic tresult;
          switch (t.type) {
            case QueryType.insert:
              if (t.table == 'DeleteLog')
                tresult = await txn.insert(t.table!, t.data!);
              else
                tresult = await txn.insert(
                    t.table!,
                    toDatabaseEntity(
                        EntityService.getEntity(t.table)!, t.data, false)!);
              break;
            case QueryType.delete:
              tresult = await txn.delete(t.table!,
                  where: t.where, whereArgs: t.whereArgs);
              break;
            case QueryType.update:
              tresult = await txn.update(
                  t.table!,
                  toDatabaseEntity(
                      EntityService.getEntity(t.table)!, t.data, true)!,
                  where: t.where,
                  whereArgs: t.whereArgs);
              break;
            case QueryType.rawQuery:
              await txn.execute(t.query!, t.whereArgs);
              break;
            default:
          }

          print('-- Result -- ${t.type}-- $tresult');
        }
      });
      return true;
    } catch (err, trace) {
      print('XXXXXXXXXXXXXXXXXXXXXXXXXXX exec trans error XXXXXXXX');
      print(err);
      print(trace);
      print(transactions.map((t) => t.type));
      return false;
    }
  }

  static Future<OperationResult> deleteEntityBulk(List<String> ids) async {
    if (!ConnectionService.isOfflineMode) return OperationResult(success: true);
    await DatabaseService.exeBatch((b) async {
      for (var id in ids) {
        b.delete('Entity', where: 'id = ?', whereArgs: [id]);
      }
      await b.commit();
      return OperationResult(success: true);
    });
    return OperationResult(success: true);
  }

  // static Future<OperationResult> insert(
  //     String table, Map<String, dynamic> data) async {
  //   if (db == null || !db.isOpen) return OperationResult(success: false);
  //   var res = await db.insert(table, data);
  //   return OperationResult(success: res > 0);
  // }

  static Future<OperationResult> writeCollection(
      String entity, List<dynamic> data) async {
    try {
      ConfigService.setValueString(entity, json.encode(data));
      // if (storage != null) {
      //   if (await storage.ready) {
      //     storage.setItem(entity, data);
      //     //storage.dispose();
      //   }
      // } else {
      //   var file = File("$pathString$entity.json");
      //   await file.writeAsString(json.encode(data), flush: true);
      // }
      return OperationResult(success: true);
    } catch (err) {
      print('XXXXXXXXXXXXXXXXX  write local collection ($entity) $err');
      return OperationResult(success: false, message: err.toString());
    }
  }

  static Future<OperationResult<List<Map<String, dynamic>>>> readCollection(
      String entity) async {
    try {
      while (!isInit) await Future.delayed(Duration(milliseconds: 100));

      var dataString = ConfigService.getValueString(entity);

      // print('--------------- read local collection ($entity)------------------');
      List<Map<String, dynamic>>? data;
      if (dataString.isNotEmpty) {
        data = (json.decode(dataString) as List?)
            ?.map<Map<String, dynamic>>((e) => e as Map<String, dynamic>)
            .toList();
      } else
        data = [];

      // if (storage != null) {
      //   //storage = new LocalStorage('localDB.json');
      //   if (await storage.ready) {
      //     data = storage.getItem(entity);
      //     //storage.dispose();
      //   }
      // } else {
      //   var file = File("$pathString$entity.json");
      //   data = json.decode(await file.readAsString());
      // }
      if (data == null) return OperationResult(success: false);
      return OperationResult(success: true, data: data);
    } catch (err) {
      print('xxx read collection error: $err');
      //print('XXXXXXXXXXXXXXXXX  read local collection ($entity) ("$pathString$entity.json") $err');
      return OperationResult(success: false, message: err.toString());
    }
  }

  static Map<String, dynamic> fromDatabaseEntity(
      Entity entity, Map<String, dynamic> databaseEntity) {
    var newData = Map<String, dynamic>();
    if (databaseEntity['id'] == null) {
      // print(
      //     'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX id = null ------fromDatabaseEntity [${entity.dbName}]');
      databaseEntity['id'] = Uuid().v4().replaceAll('-', '');
    }
    newData['id'] = databaseEntity['id'];

    if (databaseEntity['timeStamp'] != null) {
      if (databaseEntity['timeStamp'] is DateTime)
        newData['timeStamp'] =
            DateTime.fromMicrosecondsSinceEpoch(databaseEntity['timeStamp']);
    } else {
      newData['timeStamp'] = DateTime.now().toUtc();
    }

    newData['recordNumber'] = databaseEntity['recordNumber'];
    for (var prop in entity.properties) {
      if (prop.propertyType == PropertyTypes.File) {
        newData['${prop.dbName}Id'] = databaseEntity['${prop.dbName}Id'];
        continue;
      }
      if (databaseEntity[prop.dbName] == null ||
          builtInFields.contains(prop.dbName)) {
        continue;
      }
      switch (prop.valueType) {
        case ValueTypes.Boolean:
          newData[prop.dbName] = (databaseEntity[prop.dbName] == 1);
          break;
        case ValueTypes.DateTime:
          newData[prop.dbName] = DateTime.fromMillisecondsSinceEpoch(
              databaseEntity[prop.dbName],
              isUtc: true);
          break;
        case ValueTypes.Object:
          if (databaseEntity[prop.dbName] != null)
            newData[prop.dbName] = json.decode(databaseEntity[prop.dbName]);
          break;
        case ValueTypes.ObjectArray:
          if (databaseEntity[prop.dbName] != null)
            newData[prop.dbName] = json.decode(databaseEntity[prop.dbName]);
          break;
        case ValueTypes.Number:
          newData[prop.dbName] = databaseEntity[prop.dbName];
          break;
        default:
          newData[prop.dbName] = databaseEntity[prop.dbName].toString();
      }
    }
    return newData;
  }

  static dynamic getDatabaseValue(dynamic val) {
    if (val is bool)
      return val == true ? 1 : 0;
    else if (val is DateTime)
      return val.millisecondsSinceEpoch;
    else if (val is Map<String, dynamic>)
      return json.encode(val);
    else
      return val.toString();
  }

  static Map<String, dynamic>? toDatabaseEntityEx(
      Map<String, dynamic>? entityData) {
    if (entityData == null) return null;

    if (entityData['id'] == null) {
      // print(
      //     'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX id = null ------toDatabaseEntityEx [$entityData]');
      entityData['id'] = Uuid().v4().replaceAll('-', '');
    }

    if (entityData['timeStamp'] != null) {
      if (entityData['timeStamp'] is DateTime)
        entityData['timeStamp'] =
            (entityData['timeStamp'] as DateTime).millisecondsSinceEpoch;
      else {
        var date = BaseModel.parseUTC(entityData['timeStamp'].toString());
        if (date != null) entityData['timeStamp'] = date.millisecondsSinceEpoch;
      }
    } else {
      entityData['timeStamp'] = DateTime.now().toUtc().millisecondsSinceEpoch;
    }

    for (var key in entityData.keys) {
      if (entityData[key] == null) continue;
      // if (builtInFields.contains(key)) {
      //   continue;
      // }
      if (entityData[key] is bool)
        entityData[key] = entityData[key] == true ? 1 : 0;
      else if (entityData[key] is DateTime)
        entityData[key] = (entityData[key] as DateTime).millisecondsSinceEpoch;
      else if (entityData[key] is Map<String, dynamic>)
        entityData[key] = json.encode(entityData[key]);
    }
    return entityData;
  }

  static Map<String, dynamic>? toDatabaseEntity(
      Entity? entity, Map<String, dynamic>? entityData, bool forUpdate) {
    if (entity == null || entityData == null) return null;
    var newData = Map<String, dynamic>();
    if (!forUpdate && entityData['id'] == null) {
      // print(
      //     'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX id = null ------toDatabaseEntity [${entity.dbName}]');
      entityData['id'] = Uuid().v4().replaceAll('-', '');
    }
    // TODO: make sure it's ok
    if (!forUpdate) {
      newData['id'] = entityData['id'];
    }
    if (entityData['isPosted'] == null)
      newData['isPosted'] = 0;
    else if (entityData['isPosted'] is bool)
      newData['isPosted'] = (entityData['isPosted'] ?? false) ? 1 : 0;
    else if (entityData['isPosted'] is num)
      newData['isPosted'] = entityData['isPosted'] ?? 0;

    newData['postMessage'] = entityData['postMessage'];
    if (entityData['timeStampUnix'] != null &&
        entityData['timeStampUnix'] != 0) {
      newData['timeStamp'] = entityData['timeStampUnix'];
    } else if (entityData['timeStamp'] != null) {
      if (entityData['timeStamp'] is DateTime)
        newData['timeStamp'] =
            (entityData['timeStamp'] as DateTime).millisecondsSinceEpoch;
      else {
        var date = BaseModel.parseUTC(entityData['timeStamp'].toString());
        if (date != null) newData['timeStamp'] = date.millisecondsSinceEpoch;
      }
    } else {
      newData['timeStamp'] = DateTime.now().toUtc().millisecondsSinceEpoch;
    }

    if (entityData['recordNumber'] != null) {
      newData['recordNumber'] = entityData['recordNumber'];
    }

    for (var prop in entity.properties) {
      if (prop.propertyType == PropertyTypes.File) {
        newData['${prop.dbName}Id'] = entityData['${prop.dbName}Id'];
        continue;
      }
      if (builtInFields.contains(prop.dbName)) {
        continue;
      }
      switch (prop.valueType) {
        case ValueTypes.Boolean:
          newData[prop.dbName] = entityData[prop.dbName] == true ? 1 : 0;
          break;
        case ValueTypes.DateTime:
          if (entityData[prop.dbName] is DateTime)
            newData[prop.dbName] =
                (entityData[prop.dbName] as DateTime).millisecondsSinceEpoch;
          else {
            var date = BaseModel.parseUTC(entityData[prop.dbName].toString());
            if (date != null)
              newData[prop.dbName] = date.millisecondsSinceEpoch;
          }
          break;
        case ValueTypes.Object:
          if (entityData[prop.dbName] != null) {
            newData[prop.dbName] = json.encode(entityData[prop.dbName]);
          }
          break;
        case ValueTypes.ObjectArray:
          if (entityData[prop.dbName] != null) {
            newData[prop.dbName] = json.encode(entityData[prop.dbName]);
          }
          break;
        case ValueTypes.Number:
          newData[prop.dbName] = entityData[prop.dbName];
          break;
        default:
          if (entityData[prop.dbName] != null)
            newData[prop.dbName] = entityData[prop.dbName].toString();
      }
    }
    return newData;
  }
}

class DbTransaction {
  String? query;
  QueryType type;
  String? table;
  String? where;
  List<dynamic>? whereArgs;
  Map<String, dynamic>? data;
  DbTransaction(this.type,
      {this.query, this.where, this.whereArgs, this.table, this.data});
  @override
  String toString() {
    return '-- Query -- $type $table $where $whereArgs : $query $data';
  }
}

enum QueryType { insert, update, delete, rawQuery }
