part of tecfy_core_client_package;

class PorpertyFilterComponent extends StatefulWidget {
  final Property property;
  final dynamic value;
  final void Function(dynamic)? onChange;
  PorpertyFilterComponent({required this.property, this.onChange, this.value});
  @override
  _PorpertyFilterComponentState createState() =>
      _PorpertyFilterComponentState();
}

class _PorpertyFilterComponentState extends State<PorpertyFilterComponent> {
  TextEditingController? _textController;
  late FocusNode _focusNode;
  dynamic _initValue;
  bool _isVisible = true;
  @override
  void initState() {
    _init();
    super.initState();
  }

  _init() {
    _initValue = widget.value;
    if (widget.property.propertyType == PropertyTypes.Text ||
        widget.property.propertyType == PropertyTypes.Number) {
      if (_textController == null) {
        _textController =
            TextEditingController(text: widget.value?.toString() ?? '');
        _focusNode = FocusNode();
        _textController?.addListener(() {
          _submitChages();
        });
      } else {
        _textController?.text = _initValue?.toString() ?? '';
      }
    }
  }

  _submitChages() {
    if (widget.onChange == null) return;
    if (widget.property.propertyType == PropertyTypes.Text) {
      _initValue = _textController?.text;
      widget.onChange?.call(_textController?.text);
    }
    if (widget.property.propertyType == PropertyTypes.Number) {
      _initValue = num.tryParse(_textController!.text);
      widget.onChange?.call(_initValue);
    }
  }

  @override
  Widget build(BuildContext context) {
    // widget.property.isVisible(widget.data, widget.entity).then((isVisible) {
    //   if (isVisible != _isVisible) {
    //     _isVisible = isVisible;
    //     setState(() {});
    //   }
    // });

    if (!_isVisible || (!widget.property.isAdd && !widget.property.isEdit))
      return Container();

    if (_initValue != widget.value) _init();

    switch (widget.property.propertyType) {
      case PropertyTypes.Dropdown:
        return DropdownAdvancedEntity(
          (widget.property as PropertyDropdown).entityName,
          label: widget.property.name,
          entity: EntityService.getEntity(
              (widget.property as PropertyDropdown).entityName)!,
          url: (widget.property as PropertyDropdown).dropdownUrl,
          property: (widget.property as PropertyDropdown),
          isRequired: widget.property.isRequired,
          data: {widget.property.dbName: widget.value},
          onChanged: (v, data) => widget.onChange?.call(v),
        );

      case PropertyTypes.Text:
        if ((widget.property as PropertyText).isPassword) return Container();
        return AppTextFormField(
          textDirection: LocalConfigService.isArabic &&
                  !(widget.property as PropertyText).isEmail
              ? TextDirection.rtl
              : TextDirection.ltr,
          labelText: widget.property.name,
          focusNode: _focusNode,
          controller: _textController,
          onFieldSubmitted: (v) => _focusNode.unfocus(),
          textInputAction: TextInputAction.next,
          isRequired: widget.property.isRequired,
          keyboardType: (widget.property as PropertyText).isEmail
              ? TextInputType.emailAddress
              : TextInputType.text,
          maxLines: 1,
          onEditingComplete: _submitChages,
        );
      case PropertyTypes.Number:
        return AppTextFormField(
            labelText: widget.property.name,
            controller: _textController,
            focusNode: _focusNode,
            isRequired: widget.property.isRequired,
            onFieldSubmitted: (v) => _focusNode.unfocus(),
            textInputAction: TextInputAction.next,
            keyboardType: TextInputType.number,
            onEditingComplete: _submitChages);
      // case PropertyTypes.Boolean:
      //   return AppCheckBox(
      //       label: widget.property.name,
      //       initalValue: _initValue,
      //       onChanged: (v) {
      //         widget.onChange(v);
      //       });

      case PropertyTypes.Color:
        return ColorPropertyComponenet(
            initalValue: _initValue,
            isRequired: widget.property.isRequired,
            label: widget.property.name,
            onChanged: (v) => widget.onChange?.call(v));

      case PropertyTypes.Date:
        return DateProperty(
            label: widget.property.name,
            initalValue: _initValue,
            isRequired: widget.property.isRequired,
            onChanged: (v) {
              widget.onChange?.call(v);
            });

      default:
        return Container();
    }
  }
}
