part of tecfy_core_client_package;

class EntityFilterPage extends StatefulWidget {
  /// final String entityName;
  final List<Property> properties;

  /// default filters
  final List<EntityFilter>? filters;

  /// return the filters list after update
  EntityFilterPage(this.properties, [this.filters]);
  @override
  _EntityFilterPageState createState() => _EntityFilterPageState();
}

class _EntityFilterPageState
    extends AppStateView<EntityFilterPage, EntityFilterController> {
  late List<EntityFilter> filters;
  @override
  void initState() {
    _controller = EntityFilterController(this, widget.properties);
    filters = widget.filters ?? [];
    if (filters.length == 0) filters.add(EntityFilter());
    super.initState();
  }

  _onCancel() {
    Navigator.pop(context, null);
  }

  _onFilter() {
    Navigator.pop(context, filters);
  }

  _onAdd() {
    setState(() {
      filters.add(EntityFilter());
    });
  }

  @override
  Widget build(BuildContext context) {
    return Center(
      child: Container(
        constraints: BoxConstraints(
            minWidth: 300, minHeight: 300, maxHeight: 600, maxWidth: 500),
        child: Material(
          // appBar: MainAppBar(context: context,title: '${_controller.entityName} (${AppLocale.translate(context, 'Filter')})', actions: <Widget>[
          // IconButton(icon:Icon(Icons.close),onPressed: onCancel)
          // ],),
          child: DefaultTabController(
            length: 2,
            child: Column(children: [
              TabBar(labelColor: ThemeService.mainColor, tabs: [
                Tab(child: Text('Standard')),
                Tab(child: Text('Advance'))
              ]),
              Expanded(
                child: TabBarView(children: [
                  // standard
                  ListView(
                      padding: EdgeInsets.symmetric(horizontal: 10),
                      children: _controller.properties
                          .map((prop) => _getBasicFilterRow(prop))
                          .toList()),
                  // advance
                  ListView(
                    children: filters
                        .map((filter) => _getAdvanceFilterCard(filter))
                        .toList()
                      ..add(TextButton.icon(
                          icon: Icon(Icons.add),
                          onPressed: _onAdd,
                          label:
                              Text(AppLocale.translate(context, 'Add Row')))),
                  ),
                ]),
              ),
              Row(mainAxisAlignment: MainAxisAlignment.spaceEvenly, children: [
                TextButton.icon(
                    icon: Icon(Icons.search),
                    onPressed: _onFilter,
                    label: Text(AppLocale.translate(context, 'OK'))),
                TextButton.icon(
                    icon: Icon(Icons.close),
                    onPressed: _onCancel,
                    label: Text(AppLocale.translate(context, 'Close')))
              ])
            ]),
          ),
        ),
      ),
    );
  }

  Widget _getBasicFilterRow(Property prop) {
    dynamic val;
    try {
      val = filters.firstWhere((f) => f.propertyName == prop.dbName).value;
    } catch (err) {}
    return PorpertyFilterComponent(
      onChange: (v) {
        //if (v == null)
        filters.removeWhere((f) => f.propertyName == prop.dbName);
        if (v != null && v.toString().length > 0) {
          filters.add(EntityFilter(
              property: prop,
              propertyName: prop.dbName,
              operatorId: 'E',
              value: v));
        }
      },
      property: prop,
      value: val,
    );
  }

  Widget _getAdvanceFilterCard(EntityFilter filter) {
    return Card(
      margin: EdgeInsets.all(10),
      borderOnForeground: true,
      color: Colors.lightBlue[100],
      child: Column(
        children: <Widget>[
          Align(
              alignment: Alignment.topRight,
              child: IconButton(
                  icon: Icon(
                    Icons.close,
                    color: Colors.red,
                  ),
                  onPressed: () {
                    filters.remove(filter);
                    _controller.setState();
                  })),
          Row(
            children: <Widget>[
              SizedBox(width: 10),
              Expanded(
                child: _propertyList(filter),
              ),
              _operatorList(filter)
            ],
          ),
          Padding(
              child: _valueWidget(filter),
              padding: EdgeInsets.only(top: 0, bottom: 0, right: 10, left: 10))
        ],
      ),
    );
  }

  Widget _propertyList(EntityFilter filter) {
    if (_controller.properties.length == 0) return Container();
    return DropdownButtonHideUnderline(
        child: DropdownButton(
      onChanged: (v) {
        filter.property = v as Property;
        filter.value = null;

        setState(() {});
      },
      value: filter.property,
      items: _controller.properties
          .map((p) => DropdownMenuItem(child: Text(p.nameAr), value: p))
          .toList(),
    ));
  }

  DropdownButton _operatorList(EntityFilter filter) {
    return DropdownButton(
      onChanged: (v) {
        filter.operatorId = v;
        setState(() {});
      },
      value: filter.operatorId,
      items: [
        DropdownMenuItem(child: Text('='), value: 'E'),
        DropdownMenuItem(child: Text('>='), value: 'GE'),
        DropdownMenuItem(child: Text('<='), value: 'LE'),
        DropdownMenuItem(
            child: Text(AppLocale.translate(context, 'Contain')), value: 'C'),
        DropdownMenuItem(
            child: Text(AppLocale.translate(context, 'Not Contain')),
            value: 'NC'),
      ],
    );
  }

  Widget _valueWidget(EntityFilter filter) {
    if (filter.property == null)
      return Container();
    else
      return PorpertyComponent(
          property: filter.property!,
          entity: Entity.fromMap({}),
          onChange: (v) {
            setState(() => filter.value = v);
          },
          data: {filter.property?.dbName ?? '': filter.value});
  }
}
